----------------------------------------------------------------
-- Bezier curve Library by Micha Pfeiffer
----------------------------------------------------------------
-- Use however you want to, but at your own risk.


------------------------------------------------
-- Usage:
--	cPoints: List of construction points
--	res: Number of segments to draw
--	width: line width to draw the line with.
--	Example: curve = Bezier:new( {{x=10,y=20},{x=20,y=40},{x=30,y=50}}, 100, 2)
------------------------------------------------


require("Scripts/middleclass")

Bezier = class('Bezier')

local clickRadius = 15
local gridSize = 30

function Bezier:initialize( cPoints, res, width )
	self.cPoints = cPoints
	self.numCPoints = #cPoints
	res = res or 10
	self:setResolution( res )

	self.lineWidth = width or 2
end

local function factorial( num )
	local numFactorial = 1
	for i = 2, num do
		numFactorial = numFactorial * i
	end
	return numFactorial
end
local function binomCoeff( n, k )
	return factorial(n)/(factorial(k)*factorial(n-k))
end
local function dist( P1, P2 )
	return math.sqrt((P1.x-P2.x)^2 + (P1.y-P2.y)^2)
end

function Bezier:print()
	local str = "Bezier curve:\n"
	for k = 1, self.numCPoints do
		str = str .. "\t" .. "{" .. self.cPoints[k].x .. "," .. self.cPoints[k].y .. "}\n" 
	end
	print(str)
end

local function B( i, n, t )
	return binomCoeff( n, i )*(t^i)*(1-t)^(n-i)
end

function Bezier:value( t )

	local x = 0
	local y = 0
	for k = 1, #self.cPoints do
		x = x + B(k-1, self.numCPoints-1, t)*self.cPoints[k].x
		y = y + B(k-1, self.numCPoints-1, t)*self.cPoints[k].y
	end

	return { x=x, y=y }
end

function Bezier:setResolution( res )
	if not tonumber(res) then
		res = 10
	end
	self.resolution = tonumber(res)
	self:update()	-- rerender points
end

function Bezier:setLineWidth( width )
	self.lineWidth = width or 2
end

function Bezier:update()
	self.points = {}
	local step = 1/self.resolution
	local t = 0
	local prevPoint,curPoint

	for k=1,self.resolution+1 do
		t = (k-1)*step
		self.points[#self.points + 1] = self:value( t )		-- calculate point and save for drawing
	end
end

function Bezier:draw( resolution )

	love.graphics.setLineWidth( self.lineWidth )
	love.graphics.setColor(255,255,255, 255)
	for k=1,#self.points-1 do
		love.graphics.line( self.points[k].x, self.points[k].y, self.points[k+1].x, self.points[k+1].y )
	end
	love.graphics.setLineWidth( 1 )

	love.graphics.setColor(255,120,50, 150)

	if self.numCPoints >= 2 then
		--[[local startX,startY = self.cPoints[1].x, self.cPoints[1].y
		local endX, endY = self.cPoints[2].x, self.cPoints[2].y
		love.graphics.line(startX, startY, endX, endY)
		startX,startY = self.cPoints[self.numCPoints-1].x, self.cPoints[self.numCPoints-1].y
		endX, endY = self.cPoints[self.numCPoints].x, self.cPoints[self.numCPoints].y
		love.graphics.line(startX, startY, endX, endY)]]--
		for k = 1, self.numCPoints-1 do
			love.graphics.line(self.cPoints[k].x, self.cPoints[k].y, self.cPoints[k+1].x, self.cPoints[k+1].y)
		end
	end

	love.graphics.setPointStyle("rough")
	for k = 1, self.numCPoints do
		if k == self.selected then
			love.graphics.setPointSize( 6 )
			love.graphics.setColor(255,120,50, 255)
		else
			love.graphics.setPointSize( 4 )
			love.graphics.setColor(255,180,100, 255)
		end
			love.graphics.point( self.cPoints[k].x, self.cPoints[k].y )
	end
	love.graphics.setColor(255,255,255,255)
end

function Bezier:checkHit( clickPoint, ignore )		-- check if the click landed on a construction Point
	local minDist = clickRadius
	local selected
	local d
	for k,p in pairs(self.cPoints) do	-- of all points, take the one clostest to the click
		if k ~= ignore then
			d = dist( p, clickPoint )
			if d < minDist then
				selected = k
				minDist = d
			end
		end
	end
	return selected
end

function Bezier:setSelected( k )
	self.selected = k
	print("clicked point:", k)
end

function Bezier:addCPoint( P )
	self.numCPoints = self.numCPoints + 1
	self.cPoints[self.numCPoints] = P
	self:update()
	self:setSelected( self.numCPoints )		-- new points are always selected!
end

function Bezier:removeCPoint( k )
	if k < 1 or k > self.numCPoints then
		return
	end
	if self.selected == k then
		self.selected = nil
	end
	self.cPoints[k] = nil
	for c=k, self.numCPoints-1 do
		self.cPoints[c] = self.cPoints[c+1]
	end
	self.cPoints[self.numCPoints] = nil
	self.numCPoints = self.numCPoints - 1
	self:update()

	for k, p in pairs(self.cPoints) do
		print("\t" .. k, p.x, p.y)
	end
end

function Bezier:startMoving()
	self.movingPoint = true
end
function Bezier:stopMoving()
	self.movingPoint = false
end
function Bezier:isMoving()
	return self.movingPoint
end

function Bezier:movePoint( P )
	if self.selected and self.movingPoint then
		if self.snapToCPoints then
			k = self:checkHit( P, self.selected )	-- check if other buttons are close by
			if k then
				self.cPoints[self.selected].x = self.cPoints[k].x
				self.cPoints[self.selected].y = self.cPoints[k].y
			else
				if self.snapToGrid then
					self.cPoints[self.selected].x = math.floor((P.x+gridSize/2)/gridSize)*gridSize
					self.cPoints[self.selected].y = math.floor((P.y+gridSize/2)/gridSize)*gridSize
				else
					self.cPoints[self.selected].x = P.x
					self.cPoints[self.selected].y = P.y
				end
			end
		else
			if self.snapToGrid then
				self.cPoints[self.selected].x = math.floor((P.x+gridSize/2)/gridSize)*gridSize
				self.cPoints[self.selected].y = math.floor((P.y+gridSize/2)/gridSize)*gridSize
			else
				self.cPoints[self.selected].x = P.x
				self.cPoints[self.selected].y = P.y
			end
		end
		self:update()
	end
end

function Bezier:getNumCPoints()
	return self.numCPoints
end

function Bezier:setSnapToCPoints( bool )
	self.snapToCPoints = bool
end

function Bezier:setSnapToGrid( bool )
	self.snapToGrid = bool
end

